import time
import mxnet as mx
from mxnet import gluon


class Estimator:
    def __init__(self, logger, val_data, mode='image') -> None:
        self.val_data = val_data
        self.logger = logger
        self.mode = mode

    def log(self, batch_epoch, top1, top5):
        if self.mode == 'image':
            num_batch = len(self.val_data)
            self.logger.info(
                f"{batch_epoch:d} / {num_batch:d} (top1, top5) : {top1:.8f}, {top5:.8f}")
        else:
            self.logger.info(
                f"{batch_epoch:d} (top1, top5) : {top1:.8f}, {top5:.8f}")

    def test(self, network, ctx, dtype):
        acc_top1 = mx.metric.Accuracy()
        acc_top5 = mx.metric.TopKAccuracy(5)
        acc_top1.reset()
        acc_top5.reset()
        num = 0
        start = time.time()
        for i, batch in enumerate(self.val_data):
            if self.mode == 'image':
                data = gluon.utils.split_and_load(
                    batch[0], ctx_list=ctx, batch_axis=0)
                label = gluon.utils.split_and_load(
                    batch[1], ctx_list=ctx, batch_axis=0)
            else:
                data = gluon.utils.split_and_load(
                    batch.data[0], ctx_list=ctx, batch_axis=0)
                label = gluon.utils.split_and_load(
                    batch.label[0], ctx_list=ctx, batch_axis=0)
            batch_size = len(data)
            outputs = [network(X.astype(dtype, copy=False)) for X in data]
            acc_top1.update(label, outputs)
            acc_top5.update(label, outputs)

            _, top1 = acc_top1.get()
            _, top5 = acc_top5.get()
            self.log(i, top1, top5)
            num += batch_size
        end = time.time()
        speed = num / (end - start)
        self.logger.info(f'Throughput is {speed:%f} img/sec.')

        _, top1 = acc_top1.get()
        _, top5 = acc_top5.get()
        return (1-top1, 1-top5)



import logging
import os
import time
from tvm import rpc, autotvm, runtime
import vta


def get_remote(env,
               device_host="192.168.2.99",
               device_port="9091"):
    if env.TARGET not in ["sim", "tsim", "intelfocl"]:
        # 如果设置环境变量，从 tracker 节点获取 remote。
        # 要设置 tracker，您需要遵循“自动调优卷积网络用于 VTA ”教程。
        tracker_host = os.environ.get("TVM_TRACKER_HOST", None)
        tracker_port = os.environ.get("TVM_TRACKER_PORT", None)
        # 否则，如果你有设备，你想直接从 host 编程，
        # 确保你已经设置了下面的变量为你的板的 IP。
        device_host = os.environ.get("VTA_RPC_HOST", device_host)
        device_port = os.environ.get("VTA_RPC_PORT", device_port)
        # 确保使用 RPC=1 编译 TVM
        assert runtime.enabled("rpc")
        if not tracker_host or not tracker_port:
            remote = rpc.connect(device_host, int(device_port))
        else:
            remote = autotvm.measure.request_remote(
                env.TARGET,
                tracker_host,
                int(tracker_port),
                timeout=10000
            )

        # 重新配置 JIT 运行时和 FPGA。
        # 通过将路径传递给 bitstream 文件而不是 None，
        # 您可以使用自己的自定义 bitstream 编程 FPGA。
        reconfig_start = time.time()
        vta.reconfig_runtime(remote)
        vta.program_fpga(remote, bitstream=None)
        reconfig_time = time.time() - reconfig_start
        print(f"Reconfigured FPGA and RPC runtime in {reconfig_time:.2f}s!")
    # 在仿真模式中，在本地托管 RPC 服务器。
    else:
        remote = rpc.LocalSession()
        if env.TARGET in ["intelfocl"]:
            # program intelfocl aocx
            vta.program_fpga(remote, bitstream="vta.bitstream")
    return remote



from tvm.contrib.utils import TempDirectory


class TempModule(TempDirectory):
    """Create temp dir which deletes the contents when exit.

    Parameters
    ----------
    custom_path : str, optional
        Manually specify the exact temp dir path

    keep_for_debug : bool
        Keep temp directory for debugging purposes
    Returns
    -------
    temp : TempDirectory
        The temp directory object
    """
    def export_library(self, module, filename, *args, **kwargs):
        """将该 module 保存在本地临时文件夹中
        """
        path = self.relpath(filename)
        module.export_library(path, *args, **kwargs)
        return path

    def upload(self, remote, module, filename):
        """将 module 上传至 remote 并返回远程模块"""
        path = self.export_library(module, filename)
        remote.upload(path)
        # remote_module
        return remote.load_module(filename)
        


import time
from mxnet.gluon.model_zoo import vision

from tvm.ir.transform import PassContext
from tvm import relay
from tvm.contrib import utils
import vta
from vta.top import graph_pack

from .remote import get_remote


class VTAModel:
    def __init__(self, env, device, name="data"):
        self.env = env
        self.device = device
        self.remote = get_remote(self.env)
        self.lib_path = "graphlib.tar"
        self.name = name
        self.dtype_dict = {self.name: "float32"}
        

    @property
    def target(self):
        if self.device == "vta":
            return self.env.target
        else:
            return self.env.target_vta_cpu

    @property
    def ctx(self):
        if self.env.TARGET == "intelfocl":
            return [self.remote.ext_dev(0), self.remote.cpu(0)]
        else:
            # Graph runtime
            if self.device == "vta":
                return self.remote.ext_dev(0)
            else:
                return self.remote.cpu(0)

    def get_model(self, model_name):
        size = (229, 229) if model_name == "inception_v3" else (224, 224)
        shape = 3, *size
        shape_dict = {self.name: (self.env.BATCH, shape)}
        # 开始前端编译
        model = vision.get_model(model_name, pretrained=True)
        return model, shape_dict

    def compile(self, model, shape_dict, start_name, stop_name):
        target = self.target
        # 度量构建的开始时间
        build_start = time.time()
        # 开始前端编
        mod, params = relay.frontend.from_mxnet(model, shape_dict)
        # 更新 shape 和 type 字典
        shape_dict.update({k: v.shape for k, v in params.items()})
        self.dtype_dict.update({k: str(v.dtype) for k, v in params.items()})
        if target.device_name == "vta":
            # 在 Relay 中执行量化
            # 注意：为了 fold batch norm，将 `opt_level` 设置为 `3`
            with PassContext(opt_level=3):
                with relay.quantize.qconfig(global_scale=8.0, skip_conv_layers=[0]):
                    mod = relay.quantize.quantize(mod, params=params)
                # 对 VTA target 进行 graph packing 和 constant folding
                assert self.env.BLOCK_IN == self.env.BLOCK_OUT
                # 如果目标是 intelfocl 或 sim，是否有 device annotation
                relay_prog = graph_pack(
                    mod["main"],
                    self.env.BATCH,
                    self.env.BLOCK_OUT,
                    self.env.WGT_WIDTH,
                    start_name=start_name,
                    stop_name=stop_name,
                    device_annot=(self.env.TARGET == "intelfocl"),
                )
        else:
            relay_prog = mod["main"]

        # 禁用 AlterOpLayout，编译 Relay 程序
        if target.device_name != "vta":
            with PassContext(opt_level=3, disabled_pass={"AlterOpLayout"}):
                lib = relay.build(
                    relay_prog, target=target, params=params
                )
        else:
            if self.env.TARGET == "intelfocl":
                # 在 CPU 和 VTA 上运行多个目标
                target = {"cpu": self.env.target_vta_cpu,
                            "ext_dev": target}
            with vta.build_config(
                opt_level=3, disabled_pass={"AlterOpLayout", "tir.CommonSubexprElimTIR"}
            ):
                lib = relay.build(relay_prog, target=target, params=params)

        # 度量 Relay 构建时间
        build_time = time.time() - build_start
        # logging.info(f"{model} inference graph built in {build_time:.2f}s!")

        # 将 inference library 发送到远程 RPC 服务器
        temp = utils.tempdir()
        lib.export_library(temp.relpath(self.lib_path))
        self.remote.upload(temp.relpath(self.lib_path))
        lib = self.remote.load_module(self.lib_path)
        return lib, build_time, shape_dict



from dataclasses import dataclass


@dataclass
class Config:
    model: str  # 模型名称
    expected_acc: float
    batch_size: int = 1


@dataclass
class QConfig:
    model: str  # 模型名称
    expected_acc: float
    nbit_input: int = 8
    dtype_input: str = "int8"
    nbit_output: int = 16
    dtype_output: str = "int16"
    global_scale: float = 4.0
    batch_size: int = 1



"""解析 Graph JSON
"""
from typing import Any
from dataclasses import dataclass, field


@dataclass
class GraphNodeRef:
    ident: int  # 节点引用索引
    index: int = 0  # 暂不知作用
    version: int = 0  # 暂不知作用


from dataclasses import field


@dataclass
class GraphAttrs:
    """`
    Args:
        dltype: 每个节点的数据类型按顺序排列。
        device_index: 按顺序为每个节点分配设备。
        storage_id: 存储布局中每个节点的内存 slot id。
        shape: 每个节点的 k 阶形状。
        storage_id: 存储布局中每个节点的内存 slot id。
                    将参数名称映射到一对 ({storage_id: tvm.runtime.NDArray})。在运行时，可以使用 storage_id 查找参数。
    """
    dltype: list
    device_index: list
    storage_id: list
    shape: list


@dataclass
class GraphNodeAttrs:
    """
    Args:
        flatten_data: 是否需要在执行前将数据扁平化（flattened）
        func_name: 融合函数名，对应于 Relay 编译过程生成的库中的符号。
        num_inputs: 此节点的 inputs 个数
        num_outputs: 此节点产生的 outputs 个数
    """
    func_name: str
    num_inputs: str
    num_outputs: str
    flatten_data: str = "0"
    hash: str|None = None
    


@dataclass
class GraphNode:
    """
    Args:
        op: 运算类型，`null` 意味着它是占位符/变量/输入节点，`tvm_op` 意味着这个节点可以被执行
        name: 节点名字
        inputs: 运算的 inputs 位置，inputs 是包含 `(nodeid, index, version)` 的元组列表。(可选)
    """
    op: str
    name: str
    inputs: list[int] = field(default_factory=list)
    attrs: Any = None

    # def __post_init__(self):
    #     if self.op == "null":
    #         delattr(self, "attrs")


@dataclass
class GraphJson:
    """
    Args:
        arg_nodes:参数节点的索引列表，它是计算图的占位符/变量/输入节点或 constant/param。
        heads: 输出节点的索引列表。
        node_row_ptr: 存储 forward 路径的历史，所以推断任务中可以跳过某些算子来构建子图。
        attrs: 可以包含版本号或类似的有用信息。
        nodes: 节点是占位符或可计算节点。
    """
    arg_nodes: list[int]
    heads: list[GraphNodeRef]
    node_row_ptr: list[int]
    attrs: GraphAttrs
    nodes: list[GraphNode]

    def __post_init__(self):
        self.heads = [GraphNodeRef(*head) for head in self.heads]
        self.attrs = GraphAttrs(**self.attrs)
        self.nodes = [GraphNode(**node) for node in self.nodes]





import gluoncv as gcv
from mxnet import gluon
from utils.tvm import set_env
from tvm import relay
from tvm.relay import quantize as qtz


def get_model(model_name, batch_size, logger, pretrained=True, original=False, qconfig=False):
    gluon_model = gluon.model_zoo.vision.get_model(model_name, pretrained=pretrained)
    # gluon_model = gcv.model_zoo.get_model(model_name, pretrained=pretrained)
    img_size = 299 if model_name == "inceptionv3" else 224
    data_shape = (batch_size, 3, img_size, img_size)
    mod, params = relay.frontend.from_mxnet(gluon_model, {"data": data_shape})

    logger.debug("original")
    logger.debug(mod.astext(show_meta_data=False))
    if original:
        return mod, params
    with qconfig:
        logger.debug("current quantize config")
        logger.debug(qtz.current_qconfig())
        qfunc = qtz.quantize(mod, params)
        logger.debug("after quantize")
        logger.debug(qfunc.astext(show_meta_data=False))
    return qfunc, params


'''单图处理
'''
import numbers
import numpy as np
from PIL import Image


def resize(img, size, interpolation=Image.BILINEAR):
    if isinstance(size, int):
        w, h = img.size
        if (w <= h and w == size) or (h <= w and h == size):
            return img
        if w < h:
            ow = size
            oh = int(size * h / w)
            return img.resize((ow, oh), interpolation)
        else:
            oh = size
            ow = int(size * w / h)
            return img.resize((ow, oh), interpolation)
    else:
        return img.resize(size, interpolation)


def center_crop(img, output_size):
    if isinstance(output_size, numbers.Number):
        output_size = (int(output_size), int(output_size))
    w, h = img.size
    th, tw = output_size
    i = int(round((h - th) / 2.))
    j = int(round((w - tw) / 2.))
    img = img.crop((j, i, j+tw, i+th))
    return img


def mean_std_process(data, mean, std):
    mean = np.array(mean)
    mean = mean[:, np.newaxis, np.newaxis]
    std = np.array(std)
    std = std[:, np.newaxis, np.newaxis]
    data = (data-mean)/std
    return data


def process_single_image(name, mean, std):
    """对图片做预处理"""
    with Image.open(name) as im:
        im = resize(im, (256, 256))
        im = center_crop(im, 224)
        img = np.array(im)
    img = img.transpose((2, 0, 1))
    return mean_std_process(img, mean, std)

def deprocess_single_image(tensor, mean, std):
    """恢复预处理的图片"""
    img = tensor.transpose((1, 2, 0))
    mean = np.array(mean)
    std = np.array(std)
    img = (img * std) + mean
    img = img.clip(0, 255)
    return img.astype("uint8")


from mxnet import gluon

from . import set_env
import tvm
from tvm.relay import quantize as qtz
import tvm.testing
from .model import get_model
from .quantize import eval_acc


def batch_fn(batch, ctx):
    data = gluon.utils.split_and_load(batch.data[0],
                                      ctx_list=ctx, batch_axis=0)
    label = gluon.utils.split_and_load(batch.label[0],
                                       ctx_list=ctx, batch_axis=0)
    return data, label


@tvm.testing.requires_gpu
def test_acc(config, val_data, target, device,
             logger,
             log_interval=500):
    model, params = get_model(config.model,
                              config.batch_size,
                              logger,
                              original=True)
    acc = eval_acc(model, params, val_data, batch_fn,
                   target, device, logger,
                   log_interval=log_interval)
    assert acc > config.expected_acc
    return acc


@tvm.testing.requires_gpu
def test_quantize_acc(qconfig, val_data, target, device,
                      logger,
                      skip_conv_layers=[],
                      skip_dense_layer=True,
                      log_interval=500):
    tvm_qconfig = qtz.qconfig(
        skip_conv_layers=skip_conv_layers,
        skip_dense_layer=skip_dense_layer,
        nbit_input=qconfig.nbit_input,
        nbit_weight=qconfig.nbit_input,
        global_scale=qconfig.global_scale,
        dtype_input=qconfig.dtype_input,
        dtype_weight=qconfig.dtype_input,
        dtype_activation=qconfig.dtype_output,
        debug_enabled_ops=None,
    )
    model, params = get_model(qconfig.model,
                              qconfig.batch_size,
                              logger,
                              original=False,
                              qconfig=tvm_qconfig)
    acc = eval_acc(model, params, val_data, batch_fn,
                   target, device, logger,
                   log_interval=log_interval)
    assert acc > qconfig.expected_acc
    return acc



# from tvm import relay
# from tvm.ir import IRModule
# from IPython.display import display_svg
from tvm.contrib.relay_viz.dot import (
    DotPlotter,
    DotVizParser
)
from tvm_book.tvm.viz_relay import Visualizer


def graphviz_relay(ir_mod,
                   graph_name="main",
                   graph_attr={"color": "red"},
                   node_attr={"color": "blue"},
                   edge_attr={"color": "black"}):
    # 添加颜色
    dot_plotter = DotPlotter(
        graph_attr=graph_attr,
        node_attr=node_attr,
        edge_attr=edge_attr)
    viz = Visualizer(
        ir_mod,
        plotter=dot_plotter,
        parser=DotVizParser())
    return viz.display(graph_name)
    # display_svg(graph)



from dataclasses import dataclass
from pathlib import Path
from mxnet import gluon
from mxnet.gluon.data.vision import transforms
from mxnet.io import ImageRecordIter
from gluoncv.data.imagenet.classification import ImageNet
import numpy as np
from .tools import Estimator

ROOT = "/media/pc/data/4tb/lxw/remote/pc/.mxnet/datasets/imagenet"
normalize = transforms.Normalize([0.485, 0.456, 0.406],
                                 [0.229, 0.224, 0.225])


def crop(input_size, crop_ratio=0.875):
    '''仅用于验证数据集，与 TF 实现对齐，设置默认的裁剪输入比为 0.875；

    将裁剪设置为 ``ceil(input-size/ratio)``
    '''
    crop_ratio = crop_ratio if crop_ratio > 0 else 0.875
    # resize 后的尺寸
    return int(np.ceil(input_size/crop_ratio))


@dataclass
class TestsetConfig:
    batch_size: int = 32  # training batch size per device (CPU/GPU).
    input_size: int = 224  # input shape of the image
    num_workers: int = 4


def get_valset(config,
               root: str = ROOT,
               crop_ratio: float = 0.875):
    '''

    Args:
        config: TestsetConfig 实例
    '''
    resize = crop(config.input_size, crop_ratio)  # 与 TF 实现对齐
    transform_test = transforms.Compose([
        transforms.Resize(resize, keep_ratio=True),
        transforms.CenterCrop(config.input_size),
        transforms.ToTensor(),
        normalize
    ])

    imagenet_val = ImageNet(root=root, train=False)
    return gluon.data.DataLoader(
        imagenet_val.transform_first(transform_test),
        batch_size=config.batch_size,
        shuffle=False,
        num_workers=config.num_workers)


def get_valsetrec(config,
                  root: str = f'{ROOT}/rec',
                  crop_ratio: float = 0.875):
    '''

    Args:
        config: TestsetConfig 实例
        rec_dir: recio directory for validation.
    '''
    rec_dir = Path(root)
    imgrec = rec_dir/'val.rec'
    imgidx = rec_dir/'val.idx'
    resize = crop(config.input_size, crop_ratio)  # 与 TF 实现对齐
    mean_rgb = [123.68, 116.779, 103.939]
    std_rgb = [58.393, 57.12, 57.375]
    return ImageRecordIter(
        path_imgrec=imgrec,
        path_imgidx=imgidx,
        shuffle=False,
        preprocess_threads=config.num_workers,
        batch_size=config.batch_size,
        resize=resize,
        data_shape=(3, config.input_size, config.input_size),
        mean_r=mean_rgb[0],
        mean_g=mean_rgb[1],
        mean_b=mean_rgb[2],
        std_r=std_rgb[0],
        std_g=std_rgb[1],
        std_b=std_rgb[2],
    )


def imagenet_estimator(logger, config, mode='image',
                       crop_ratio: float = 0.875):
    root = ROOT if mode == 'image' else f'{ROOT}/rec'
    kwargs = {
        'config': config,
        'crop_ratio': crop_ratio,
        'root': root
    }
    if mode == 'image':
        val_data = get_valset(**kwargs)
    else:
        val_data = get_valsetrec(**kwargs)
    return Estimator(logger, val_data, mode)



# This file is generated automatically through:
#    d2lbook build lib
# Don't edit it directly
# 来源于 http://tvm.d2l.ai/chapter_getting_started/install.html

import sys
d2ltvm = sys.modules[__name__]

# Defined in file: ./chapter_getting_started/install.md
import tvm
from tvm import te
import time
import timeit
import numpy as np
from matplotlib import pyplot as plt
from IPython import display
try:
    import mxnet as mx
except:
    pass


# Defined in file: ./chapter_getting_started/vector_add.md
def get_abc(shape, constructor=None):
    """Return random a, b and empty c with the same shape.
    """
    np.random.seed(0)
    a = np.random.normal(size=shape).astype(np.float32)
    b = np.random.normal(size=shape).astype(np.float32)
    c = np.empty_like(a)
    if constructor:
        a, b, c = [constructor(x) for x in (a, b, c)]
    return a, b, c


# Defined in file: ./chapter_getting_started/vector_add.md
def vector_add(n):
    """TVM expression for vector add"""
    A = te.placeholder((n,), name='a')
    B = te.placeholder((n,), name='b')
    C = te.compute(A.shape, lambda i: A[i] + B[i], name='c')
    return A, B, C


# Defined in file: ./chapter_getting_started/from_mxnet.md
def image_preprocessing(image):
    image = np.array(image) - np.array([123., 117., 104.])
    image /= np.array([58.395, 57.12, 57.375])
    image = image.transpose((2, 0, 1))
    image = image[np.newaxis, :]
    return image.astype('float32')


# Defined in file: ./chapter_common_operators/broadcast_add.md
def broadcast_add(shape1, shape2):
    """Broadcast add between two 2-dimensional tensors

    shape1, shape2 : the shapes of the input tensors
    """
    assert len(shape1) == 2 and len(shape2) == 2, \
        "broadcast tensors should both be 2-dimension"
    for i in range(len(shape1)):
        assert shape1[i] == shape2[i] or shape1[i] == 1 or shape2[i] == 1, \
            "tensor shapes do not fit for broadcasting"
    A = te.placeholder(shape1, name='A')
    B = te.placeholder(shape2, name='B')
    m = shape1[0] if shape2[0] == 1 else shape2[0]
    n = shape1[1] if shape2[1] == 1 else shape2[1]
    f = lambda x, y: A[0 if shape1[0]==1 else x, 0 if shape1[1]==1 else y] + \
        B[0 if shape2[0]==1 else x, 0 if shape2[1]==1 else y]
    C = te.compute((m, n), f, name='C')
    return A, B, C


# Defined in file: ./chapter_common_operators/broadcast_add.md
def get_bcast_data(shape1, shape2, constructor=None):
    """Return random tensors a, b
    and empty tensor c to store broadcast results between a and b

    shape1, shape2: shapes of input tensors
    constructor : user-defined tensor constructor
    """
    np.random.seed(0)
    a = np.random.normal(size=shape1).astype("float32")
    b = np.random.normal(size=shape2).astype("float32")
    out_shape = (shape1[0] if shape2[0] == 1 else shape2[0],
                 shape1[1] if shape2[1] == 1 else shape2[1])
    c = np.empty(out_shape, dtype='float32')
    if constructor:
        a, b, c = [constructor(x) for x in (a, b, c)]
    return a, b, c


# Defined in file: ./chapter_common_operators/matmul.md
def matmul(n, m, l):
    """Return the computing expression of matrix multiplication
    A : n x l matrix
    B : l x m matrix
    C : n x m matrix with C = A B
    """
    k = te.reduce_axis((0, l), name='k')
    A = te.placeholder((n, l), name='A')
    B = te.placeholder((l, m), name='B')
    C = te.compute((n, m),
                    lambda x, y: te.sum(A[x, k] * B[k, y], axis=k),
                    name='C')
    return A, B, C


# Defined in file: ./chapter_common_operators/conv.md
def padding(X, ph, pw, val=0):
    """Pad X with the given value in 2-D

    ph, pw : height and width padding
    val : padding value, default 0
    """
    assert len(X.shape) >= 2
    nh, nw = X.shape[-2], X.shape[-1]
    return te.compute(
            (*X.shape[0:-2], nh+ph*2, nw+pw*2),
            lambda *i: te.if_then_else(
                te.any(i[-2]<ph, i[-2]>=nh+ph, i[-1]<pw, i[-1]>=nw+pw),
                val, X[i[:-2]+(i[-2]-ph, i[-1]-pw)]),
            name='PaddedX')


# Defined in file: ./chapter_common_operators/conv.md
def conv_out_size(n, k, p, s):
    """Compute the output size by given input size n (width or height),
    kernel size k, padding p, and stride s
    Return output size (width or height)
    """
    return (n - k + 2 * p)//s + 1


# Defined in file: ./chapter_common_operators/conv.md
def conv(oc, ic, nh, nw, kh, kw, ph=0, pw=0, sh=1, sw=1):
    """Convolution

    oc, ic : output and input channels
    nh, nw : input width and height
    kh, kw : kernel width and height
    ph, pw : height and width padding sizes, default 0
    sh, sw : height and width strides, default 1
    """
    # reduction axes
    ric = te.reduce_axis((0, ic), name='ric')
    rkh = te.reduce_axis((0, kh), name='rkh')
    rkw = te.reduce_axis((0, kw), name='rkw')
    # output height and weights
    oh = conv_out_size(nh, kh, ph, sh)
    ow = conv_out_size(nw, kw, pw, sw)
    # pad X and then compute Y
    X = te.placeholder((ic, nh, nw), name='X')
    K = te.placeholder((oc, ic, kh, kw), name='K')
    PaddedX = padding(X, ph, pw) if ph * pw != 0 else X
    Y = te.compute(
        (oc, oh, ow),
        lambda c, i, j: te.sum(
            PaddedX[ric, i*sh+rkh, j*sw+rkw] * K[c, ric, rkh, rkw],
            axis=[ric, rkh, rkw]), name='Y')
    return X, K, Y, PaddedX


# Defined in file: ./chapter_common_operators/conv.md
def conv_mxnet(data, weight, bias, out, k, p, s):
    mx.nd.Convolution(data, weight, bias, kernel=(k,k), stride=(s,s),
                      pad=(p,p), num_filter=out.shape[1], out=out)


# Defined in file: ./chapter_common_operators/depthwise_conv.md
def get_conv_data(oc, ic, n, k, p=0, s=1, constructor=None, conv_type='direct'):
    """Return random 3-D data tensor, 3-D kernel tenor and empty 3-D output
    tensor with the shapes specified by input arguments.

    oc, ic : output and input channels
    n : input width and height
    k : kernel width and height
    p : padding size, default 0
    s : stride, default 1
    conv_type: either direct 2D or depthwise, default direct
    constructor : user-defined tensor constructor
    """
    np.random.seed(0)
    data = np.random.normal(size=(ic, n, n)).astype('float32')
    ic_weight = ic
    if conv_type == 'depthwise':
        ic_weight = 1
    weight = np.random.normal(size=(oc, ic_weight, k, k)).astype('float32')
    on = d2ltvm.conv_out_size(n, k, p, s)
    out = np.empty((oc, on, on), dtype='float32')
    if constructor:
        data, weight, out = (constructor(x) for x in [data, weight, out])
    return data, weight, out


# Defined in file: ./chapter_common_operators/depthwise_conv.md
def depthwise_conv(ic, nh, nw, kh, kw, ph=0, pw=0, sh=1, sw=1):
    """Convolution

    ic : number of channels for both input and output
    nh, nw : input width and height
    kh, kw : kernel width and height
    ph, pw : height and width padding sizes, default 0
    sh, sw : height and width strides, default 1
    """
    # reduction axes
    rkh = te.reduce_axis((0, kh), name='rkh')
    rkw = te.reduce_axis((0, kw), name='rkw')
    # output height and weights
    oh = conv_out_size(nh, kh, ph, sh)
    ow = conv_out_size(nw, kw, pw, sw)
    # pad X and then compute Y
    X = te.placeholder((ic, nh, nw), name='X')
    K = te.placeholder((ic, 1, kh, kw), name='K')
    PaddedX = padding(X, ph, pw) if ph * pw != 0 else X
    Y = te.compute(
        (ic, oh, ow),
        lambda c, i, j: te.sum(
            (PaddedX[c, i*sh+rkh, j*sw+rkw] * K[c, 0, rkh, rkw]),
            axis=[rkh, rkw]), name='Y')

    return X, K, Y, PaddedX


# Defined in file: ./chapter_common_operators/depthwise_conv.md
def get_conv_data_mxnet(oc, ic, n, k, p, s, ctx='cpu', conv_type='direct'):
    ctx = getattr(mx, ctx)()
    data, weight, out = get_conv_data(oc, ic, n, k, p, s,
                                      constructor=lambda x: mx.nd.array(x, ctx=ctx),
                                      conv_type=conv_type)
    data, out = data.expand_dims(axis=0), out.expand_dims(axis=0)
    bias = mx.nd.zeros(out.shape[1], ctx=ctx)
    return data, weight, bias, out


# Defined in file: ./chapter_common_operators/depthwise_conv.md
def depthwise_conv_mxnet(data, weight, bias, out, k, p, s):
    mx.nd.Convolution(data, weight, bias, kernel=(k,k), stride=(s,s),
                      pad=(p,p), num_filter=out.shape[1],
                      out=out, num_group=weight.shape[0])


# Defined in file: ./chapter_common_operators/pooling.md
def pool(pool_type, c, nh, nw, kh, kw, ph=0, pw=0, sh=1, sw=1):
    """2D pooling

    pool_type: pooling type, 'max' or 'avg'
    c : channels
    nh, nw : input width and height
    kh, kw : kernel width and height
    ph, pw : height and width padding sizes, default 0
    sh, sw : height and width strides, default 1
    """
    # reduction axes
    rkh = te.reduce_axis((0, kh), name='rkh')
    rkw = te.reduce_axis((0, kw), name='rkw')
    # output height and weights
    oh = d2ltvm.conv_out_size(nh, kh, ph, sh)
    ow = d2ltvm.conv_out_size(nw, kw, pw, sw)
    # pad X and then compute Y
    X = te.placeholder((c, nh, nw), name='X')


    if pool_type == 'max':
        PaddedX = d2ltvm.padding(X, ph, pw, val=te.min_value(X.dtype)) \
            if ph * pw != 0 else X
        Y = te.compute((c, oh, ow), \
                            lambda c, h, w: \
                            te.max(PaddedX[c, h*sh+rkh, w*sw+rkw], \
                                axis=[rkh, rkw]), \
                            tag="pool_max", name='PoolMax')
    elif pool_type == 'avg':
        PaddedX = d2ltvm.padding(X, ph, pw) if ph * pw != 0 else X
        tsum = te.compute((c, oh, ow), \
                            lambda c, h, w: \
                            te.sum(PaddedX[c, h*sh+rkh, w*sw+rkw], \
                                axis=[rkh, rkw]), \
                            tag="pool_avg1", name='PoolSum')
        Y = te.compute((c, oh, ow), \
                            lambda c, h, w: \
                            tsum[c, h, w] / (kh*kw), \
                            tag='pool_avg2', name='PoolAvg')
    else:
        raise ValueError("Pool type should be 'avg' or 'max'.")
    return X, Y, PaddedX


# Defined in file: ./chapter_common_operators/pooling.md
def get_pool_data_mxnet(c, n, k, p, s, ctx='cpu'):
    ctx = getattr(mx, ctx)()
    data, _, out = d2ltvm.get_conv_data(c, c, n, k, p, s,
                                      lambda x: mx.nd.array(x, ctx=ctx))
    data, out = data.expand_dims(axis=0), out.expand_dims(axis=0)
    return data, out


# Defined in file: ./chapter_common_operators/pooling.md
def pool_mxnet(pool_type, data, out, k, p, s):
    mx.nd.Pooling(data, kernel=(k,k), stride=(s,s),
                      pad=(p,p), pool_type=pool_type, out=out)


# Defined in file: ./chapter_common_operators/batch_norm.md
from tvm import topi

def batch_norm(c, n, eps=1e-5):
    """batch normalization

    c : channels
    N : input width and height
    eps : small positive value to prevent divide 0
    """

    X = te.placeholder((c, n, n), name='X')
    Mean = te.placeholder((c, 1, 1), name='Mean')
    Var = te.placeholder((c, 1, 1), name='Var')
    Gamma = te.placeholder((c, 1, 1), name='Gamma')
    Beta = te.placeholder((c, 1, 1), name='Beta')
    C1 = X - Mean
    C2 = topi.sqrt(Var + eps)
    Y = C1 / C2 * Gamma + Beta
    return X, Mean, Var, Gamma, Beta, Y


# Defined in file: ./chapter_common_operators/batch_norm.md
def get_bn_data(c, n, constructor=None):
    """Return the batch norm data, mean, variance, gamma and beta tensors.
       Also return the empty tensor for output.

    c : channels
    n : input width and height
    constructor : user-defined tensor constructor
    """
    np.random.seed(0)
    data = np.random.normal(size=(c, n, n)).astype('float32')
    mean = np.random.normal(size=(c, 1, 1)).astype('float32')
    # move the mean of the normal distribution to be 1
    var = np.random.normal(loc=1.0, size=(c, 1, 1)).astype('float32')
    # make sure all variance numbers are not negative
    var = np.absolute(var)
    gamma = np.random.normal(size=(c, 1, 1)).astype('float32')
    beta = np.random.normal(size=(c, 1, 1)).astype('float32')
    out = np.empty((c, n, n), dtype='float32')
    if constructor:
        data, mean, var, gamma, beta, out = \
        (constructor(x) for x in [data, mean, var, gamma, beta, out])
    return data, mean, var, gamma, beta, out


# Defined in file: ./chapter_common_operators/batch_norm.md
def get_bn_data_mxnet(c, n, ctx='cpu'):
    ctx = getattr(mx, ctx)()
    data, mean, var, gamma, beta, out = get_bn_data(c, n,
                                      lambda x: mx.nd.array(x, ctx=ctx))
    data, out = data.expand_dims(axis=0), out.expand_dims(axis=0)
    return data, mean, var, gamma, beta, out


# Defined in file: ./chapter_common_operators/batch_norm.md
def batch_norm_mxnet(data, mean, var, gamma, beta, out, eps=1e-5):
    # use_global_stats=True to use the input mean and var instead of computing
    # the mean and var of the input data.
    # fix_gamma=False so that gamma won't be set to 1.
    mx.nd.BatchNorm(data, gamma, beta, mean, var, eps,
                    use_global_stats=True, fix_gamma=False, out=out)


# Defined in file: ./chapter_cpu_schedules/call_overhead.md
def bench_workload(workload):
    """Benchmark a workload

    workload: a method that accept a num_repeat argument
    and return its total execution time
    """
    workload(1)  # warmup
    time = workload(1)  # the time to run once
    if time > 1: return time
    # The number of repeats to measure at least 1 second
    num_repeats = max(int(1.0 / time), 5)
    return workload(num_repeats) / num_repeats


# Defined in file: ./chapter_cpu_schedules/vector_add.md
def plot(X, Y, xlabel=None, ylabel=None, legend=[], xlim=None,
         ylim=None, xscale='linear', yscale='linear', fmts=None,
         figsize=(4.5, 3)):
    """Plot multiple lines"""
    display.set_matplotlib_formats('svg')
    plt.rcParams['figure.figsize'] = figsize
    axes = plt.gca()
    X, Y = np.array(X), np.array(Y)
    if X.shape != Y.shape: X = [X] * len(Y)
    if not fmts: fmts = ['-'] * len(X)
    for x, y, fmt in zip(X, Y, fmts):
        axes.plot(x, y, fmt)
    axes.set_xlabel(xlabel)
    axes.set_ylabel(ylabel)
    axes.set_xscale(xscale)
    axes.set_yscale(yscale)
    axes.set_xlim(xlim)
    axes.set_ylim(ylim)
    if legend: axes.legend(legend)
    axes.grid()


# Defined in file: ./chapter_cpu_schedules/vector_add.md
def plot_gflops(sizes, gflops, legend, xlabel='Size'):
    d2ltvm.plot(sizes, gflops, xlabel=xlabel, ylabel='GFLOPS',
             xscale='log', yscale='log',
             legend=legend, fmts=['--']*(len(gflops)-1)+['-'])


# Defined in file: ./chapter_cpu_schedules/vector_add.md
def bench_vector_add_tvm(func, sizes, target):
    def workload(nrepeats):
        timer = mod.time_evaluator(mod.entry_name, ctx=ctx, number=nrepeats)
        return timer(a, b, c).mean * nrepeats
    times = []
    for n in sizes:
        s, (A, B, C) = func(int(n))
        mod = tvm.build(s, [A, B, C], target)
        ctx = tvm.context(target, 0)
        a, b, c = d2ltvm.get_abc(n, lambda x: tvm.nd.array(x, ctx=ctx))
        times.append(d2ltvm.bench_workload(workload))
    return sizes / 1e9 / np.array(times)


# Defined in file: ./chapter_cpu_schedules/broadcast_add.md
def bench_bcast_add_tvm(func, sizes, target):
    def workload(nrepeats):
        timer = mod.time_evaluator(mod.entry_name, ctx=ctx, number=nrepeats)
        return timer(a, b, c).mean * nrepeats
    times = []
    for n in sizes:
        n = int(n)
        s, (A, B, C) = func(n)
        mod = tvm.build(s, [A, B, C], target)
        ctx = tvm.context(target, 0)
        a, b, c = d2ltvm.get_bcast_data((n, 1), (n, n), lambda x: tvm.nd.array(x, ctx=ctx))
        times.append(d2ltvm.bench_workload(workload))
    return sizes * sizes / 1e9 / np.array(times)


# Defined in file: ./chapter_cpu_schedules/matmul.md
def np_matmul_timer(n):
    timer = timeit.Timer(setup='import numpy as np\n'
                         'import d2ltvm\n'
                         'a, b, c = d2ltvm.get_abc(%s)' % str((n,n)),
                         stmt = 'np.dot(a, b, out=c)')
    return timer.timeit


# Defined in file: ./chapter_cpu_schedules/matmul.md
def bench_matmul_tvm(func, sizes, target):
    def workload(nrepeats):
        timer = mod.time_evaluator(mod.entry_name, ctx=ctx, number=nrepeats)
        return timer(a, b, c).mean * nrepeats
    times = []
    for n in sizes:
        s, (A, B, C) = func(int(n))
        mod = tvm.build(s, [A, B, C], target)
        ctx = tvm.context(target, 0)
        a, b, c = d2ltvm.get_abc((n, n), lambda x: tvm.nd.array(x, ctx=ctx))
        times.append(d2ltvm.bench_workload(workload))
    return 2 * sizes**3 / 1e9 / np.array(times)


# Defined in file: ./chapter_cpu_schedules/conv.md
def conv_gflop(oc, ic, n, k, p, s):
    """Compute the #floating point operations in a convolution.

    The arguments are output channels oc, input channels ic, input size n,
    kernel size k, padding p and stride s.
    """
    on = d2ltvm.conv_out_size(n, k, p, s)
    return 2 * oc * ic * on * on * k * k / 1e9


# Defined in file: ./chapter_cpu_schedules/conv.md
def conv_timer_mxnet(c, n, k, ctx):
    """Benchmark convolution in MXNet

    c : input, output channels
    n : input width and height
    k : kernel width and height
    """
    timer = timeit.Timer(
        setup='import d2ltvm\n'
        'import mxnet as mx\n'
        'c, n, k, p, s = %d, %d, %d, %d, 1\n'
        'data, weight, bias, out = d2ltvm.get_conv_data_mxnet(\n'
        '    c, c, n, k, p, s, "%s")'%(c, n, k, (k-1)//2, ctx),
        stmt='d2ltvm.conv_mxnet(data, weight, bias, out, k, p, s);'
        'out.wait_to_read()')
    return timer.timeit


# Defined in file: ./chapter_cpu_schedules/conv.md
def bench_conv_mxnet(sizes, ctx='cpu'):
    """Return the GFLOPS of MXNet convolution"""
    return [conv_gflop(c, c, n, k, (k-1)//2, 1) /
            d2ltvm.bench_workload(conv_timer_mxnet(c, n, k, ctx))
            for c, n, k in sizes]


# Defined in file: ./chapter_cpu_schedules/conv.md
def bench_conv_tvm(func, sizes, target):
    def workload(nrepeats):
        timer = mod.time_evaluator(mod.entry_name, ctx=ctx, number=nrepeats)
        return timer(x, k, y).mean * nrepeats
    gflops, times = [], []
    for (c, n, k) in sizes:
        args = c, c, n, k, (k-1)//2, 1 # oc, ic, n, k, p, s
        s, (X, K, Y) = func(*args)
        mod = tvm.build(s, [X, K, Y], target)
        ctx = tvm.context(target, 0)
        x, k, y = d2ltvm.get_conv_data(
            *args, lambda x: tvm.nd.array(x, ctx=ctx))
        times.append(d2ltvm.bench_workload(workload))
        gflops.append(conv_gflop(*args))
    return np.array(gflops) / np.array(times)


# Defined in file: ./chapter_cpu_schedules/depthwise_conv.md
def bench_depthwise_conv_tvm(func, sizes, target):
    def workload(nrepeats):
        timer = mod.time_evaluator(mod.entry_name, ctx=ctx, number=nrepeats)
        return timer(x, k, y).mean * nrepeats
    gflops, times = [], []
    for (c, n, k) in sizes:
        args = c, n, k, (k-1)//2, 1 # c, n, k, p, s
        s, (X, K, Y) = func(*args)
        mod = tvm.build(s, [X, K, Y], target)
        ctx = tvm.context(target, 0)
        x, k, y = d2ltvm.get_conv_data(
            args[0], *args, lambda x: tvm.nd.array(x, ctx=ctx), conv_type='depthwise')
        times.append(d2ltvm.bench_workload(workload))
        gflops.append(d2ltvm.conv_gflop(1, *args))
    return np.array(gflops) / np.array(times)


# Defined in file: ./chapter_cpu_schedules/depthwise_conv.md
def depthwise_conv_timer_mxnet(c, n, k, ctx):
    """Benchmark convolution in MXNet

    c : input, output channels
    n : input width and height
    k : kernel width and height
    """
    timer = timeit.Timer(
        setup='import d2ltvm\n'
        'import mxnet as mx\n'
        'c, n, k, p, s = %d, %d, %d, %d, 1\n'
        'data, weight, bias, out = d2ltvm.get_conv_data_mxnet(\n'
        '    c, c, n, k, p, s, "%s", "%s")'%(c, n, k, (k-1)//2, ctx, 'depthwise'),
        stmt='d2ltvm.depthwise_conv_mxnet(data, weight, bias, out, k, p, s);'
        'out.wait_to_read()')
    return timer.timeit


# Defined in file: ./chapter_cpu_schedules/depthwise_conv.md
def bench_depthwise_conv_mxnet(sizes, ctx='cpu'):
    """Return the GFLOPS of MXNet convolution"""
    return [d2ltvm.conv_gflop(1, c, n, k, (k-1)//2, 1) /
            d2ltvm.bench_workload(depthwise_conv_timer_mxnet(c, n, k, ctx))
            for c, n, k in sizes]


# Defined in file: ./chapter_cpu_schedules/pooling.md
def bench_pooling_tvm(func, sizes, target):
    """Benchmark pooling in TVM

    func : the scheduling method
    sizes : the data size list, each of which is a (channel, input_hw, kernel_hw) triplet
    target : the TVM target, e.g. llvm or cuda
    """
    def workload(nrepeats):
        timer = mod.time_evaluator(mod.entry_name, ctx=ctx, number=nrepeats)
        return timer(data, out_max).mean * nrepeats
    times = []
    for size in sizes:
        sch, args = func(size)
        mod = tvm.build(sch, args, target)
        ctx = tvm.context(target, 0)
        data, _, out_max = d2ltvm.get_conv_data(size[0], size[0], size[1], size[2], 1, 1,
                                                lambda x: tvm.nd.array(x, ctx=ctx))
        times.append(d2ltvm.bench_workload(workload))
    return np.array(times)


# Defined in file: ./chapter_cpu_schedules/pooling.md
def pooling_timer_mxnet(pool_type, c, n, k, ctx):
    """Benchmark pooling in MXNet

    c : channels
    n : input width and height
    k : kernel width and height
    """
    timer = timeit.Timer(
        setup='import d2ltvm\n'
        'import mxnet as mx\n'
        'c, n, k, p, s = %d, %d, %d, 1, 1\n'
        'data, out = d2ltvm.get_pool_data_mxnet(\n'
        '    c, n, k, p, s, "%s")'%(c, n, k, ctx),
        stmt='d2ltvm.pool_mxnet("%s", data, out, k, p, s);'
        'out.wait_to_read()'%(pool_type))
    return timer.timeit


# Defined in file: ./chapter_cpu_schedules/pooling.md
def bench_pooling_mxnet(pool_type, sizes, ctx='cpu'):
    """Return the execution times of MXNet pooling"""
    return [d2ltvm.bench_workload(pooling_timer_mxnet(pool_type, c, n, k, ctx))
            for c, n, k in sizes]


# Defined in file: ./chapter_cpu_schedules/batch_norm.md
def bench_bn_tvm(func, sizes, target):
    """Benchmark batch normalization in TVM

    func : the scheduling method
    sizes : the data size list, each of which is a (channel, input_hw) tuple
    target : the TVM target, e.g. llvm or cuda
    """
    def workload(nrepeats):
        timer = mod.time_evaluator(mod.entry_name, ctx=ctx, number=nrepeats)
        return timer(data, mean, var, gamma, beta, out).mean * nrepeats
    times = []
    for size in sizes:
        sch, args = func(size)
        mod = tvm.build(sch, args, target)
        ctx = tvm.context(target, 0)
        data, mean, var, gamma, beta, out = d2ltvm.get_bn_data(size[0], size[1],
                                                               lambda x: tvm.nd.array(x, ctx=ctx))
        times.append(d2ltvm.bench_workload(workload))
    return np.array(times)


# Defined in file: ./chapter_cpu_schedules/batch_norm.md
def bn_timer_mxnet(c, n, ctx):
    """Benchmark batch normalization in MXNet

    c : channels
    n : input width and height
    ctx : compute ctx, e.g., cpu or gpu
    """
    timer = timeit.Timer(
        setup='import d2ltvm\n'
        'import mxnet as mx\n'
        'c, n = %d, %d\n'
        'data, mean, var, gamma, beta, out = d2ltvm.get_bn_data_mxnet(\n'
        '    c, n, "%s")'%(c, n, ctx),
        stmt='d2ltvm.batch_norm_mxnet(data, mean, var, gamma, beta, out);'
        'out.wait_to_read()')
    return timer.timeit


# Defined in file: ./chapter_cpu_schedules/batch_norm.md
def bench_bn_mxnet(sizes, ctx='cpu'):
    """Return the execution times of MXNet batch norm"""
    return [d2ltvm.bench_workload(bn_timer_mxnet(c, n, ctx))
            for c, n in sizes]


# Defined in file: ./chapter_gpu_schedules/matmul.md
def matmul_timer_mxnet(n, ctx):
    """The matrix multiplication timer for MXNet

    n : width and height of inputs
    ctx : device
    """
    timer = timeit.Timer(
        setup='import d2ltvm\n'
        'import mxnet as mx\n'
        'a, b, c, = d2ltvm.get_abc((%d, %d), lambda x: mx.nd.array(x, ctx=mx.%s()))\n'
        'mx.nd.waitall()' % (n, n, ctx),
        stmt='mx.nd.dot(a, b, out=c); c.wait_to_read()')
    return timer.timeit


# Defined in file: ./chapter_gpu_schedules/conv.md
def split_axis(factors, sch, op, axis):
        """Splitting an axis into factors

        Parameters
        ----------
        factors: array of integers
            The factors that the split applies
        sch: tvm.te.schedule.Schedule
            The tvm schedule
        op: tvm.te.tensor.Operation
            The stage to be applied
        axis: tvm.te.schedule.IterVar
            axis to split

        Returns
        -------
        axes : list of Axis
            The transformed axes.
        """
        ret = []
        for i in range(0, len(factors)):
            ax0, ax1 = sch[op].split(axis, factor=int(np.prod(factors[i:])))
            ret.append(ax0)
            axis = ax1
        return ret + [axis]





from graphviz import Digraph
from tvm.contrib.relay_viz import RelayVisualizer


class Visualizer(RelayVisualizer):
    def graph(self, graph_name):
        return self._plotter._name_to_graph[graph_name]

    def display(self, graph_name):
        graph = self.graph(graph_name)
        return graph.digraph

    def display_all(self, format="svg",
                    filename=None,
                    directory="images"):
        root_graph = Digraph(format=format,
                             filename=filename,
                             directory=directory)
        for graph in self._plotter._name_to_graph.values():
            root_graph.subgraph(graph.digraph)
        return root_graph


if __name__ == "__main__":
    from tvm import relay
    from tvm.ir import IRModule
    from IPython.display import display_svg
    from tvm.contrib.relay_viz.dot import (
        DotPlotter,
        DotVizParser
    )
    from tvm_book.tvm.viz_relay import Visualizer


    x = relay.var("x")
    sb = relay.ScopeBuilder()
    v1 = sb.let("v1", relay.log(x))
    v2 = sb.let("v2", v1 + v1)
    sb.ret(v2)
    f = relay.Function([x], sb.get())

    mod = IRModule()
    mod["main"] = f

    graph_attr = {"color": "red"}
    node_attr = {"color": "blue"}
    edge_attr = {"color": "black"}

    # 添加颜色
    dot_plotter = DotPlotter(
        graph_attr=graph_attr,
        node_attr=node_attr,
        edge_attr=edge_attr)
    viz = Visualizer(
        mod,
        plotter=dot_plotter,
        parser=DotVizParser())
    graph = viz.display("main")
    display_svg(graph)


import sys
import os
from pathlib import Path
from importlib import import_module


def set_tvm(tvm_root):
    tvm_root = Path(tvm_root)
    TVM_PATH = str(tvm_root/'python')
    VTA_PATH = str(tvm_root/'vta/python')
    # sys.path.extend([TVM_PATH, VTA_PATH])
    for path in [TVM_PATH, VTA_PATH]:
        if path not in sys.path:
            sys.path.extend([path])
    vta_path = str(tvm_root/"3rdparty/vta-hw")
    os.environ['VTA_HW_PATH'] = os.environ.get('VTA_HW_PATH', vta_path)


def set_mxnet():
    os.environ['MXNET_CUDNN_AUTOTUNE_DEFAULT'] = '0'
    os.environ['MXNET_CUDNN_LIB_CHECKING'] = '0'


def set_cudnn(cuda_path=":/usr/local/cuda/bin",
              LD_LIBRARY_PATH="/usr/local/cuda/lib64"):
    os.environ["PATH"] += cuda_path
    os.environ["LD_LIBRARY_PATH"] = LD_LIBRARY_PATH


def import_tvm(tvm_root):
    set_tvm(tvm_root)
    # import 模块
    tvm = import_module('tvm')
    vta = import_module('vta')
    return tvm, vta



import mxnet as mx

from . import set_env
import tvm
from tvm import relay


def eval_acc(model, params,
             dataset, batch_fn,
             target, device,
             logger,
             log_interval=500):
    with tvm.transform.PassContext(opt_level=3):
        lib = relay.build(model, target, params=params)
    # create runtime module
    m = tvm.contrib.graph_executor.GraphModule(lib["default"](device))

    # setup evaluaiton metric
    if isinstance(dataset, mx.io.io.MXDataIter):
        dataset.reset()
        batch_size = dataset.batch_size
    else:
        batch_size = 1
        
    acc_top1 = mx.metric.Accuracy()
    acc_top5 = mx.metric.TopKAccuracy(5)
    acc_top1.reset()
    acc_top5.reset()
    # Execute
    for i, batch in enumerate(dataset):
        data, label = batch_fn(batch, [mx.cpu(0)])
        m.set_input("data", tvm.nd.array(data[0].asnumpy()))
        m.run()
        out_arr = m.get_output(0)
        acc_top1.update(label, [mx.nd.array(out_arr.numpy())])
        acc_top5.update(label, [mx.nd.array(out_arr.numpy())])

        if not (i + 1) % log_interval:
            _, top1 = acc_top1.get()
            _, top5 = acc_top5.get()
            nsamples = (i + 1) * batch_size
            logger.info(
                "[%d samples] validation: acc-top1=%f acc-top5=%f", nsamples, top1, top5)
    logger.info("[final] validation: acc-top1=%f acc-top5=%f", top1, top5)
    return top1





import numpy as np


def pad2d(X, ph, pw, val=0):
    """Pad X with the given value in 2-D

    ph, pw : height and width padding
    val : padding value, default 0
    """
    assert len(X.shape) >= 2
    nh, nw = X.shape[2], X.shape[3]
    pad_shape = *X.shape[:2], nh+ph*2, nw+pw*2, *X.shape[4:]
    padX = np.ones(pad_shape, X.dtype) * val
    padX[:, :, ph:-ph, pw:-pw] = X
    return padX



from typing import NamedTuple
import numpy as np
from .padding import pad2d


def conv_out_size(n, k, p, s):
    """根据给定的输入大小计算输出大小，
    
    Args:
        n: 宽或高
        k: kernel 大小
        p: 填充
        s: 步幅

    Returns:
        输出大小(宽度或高度)
    """
    return (n - k + 2 * p)//s + 1


class Conv2d(NamedTuple):
    """根据 packing 推导形状

    Args:
        batch: 数据量
        height: 高度
        width: 宽度
        in_channels: 输入通道
        out_channels: 输出通道
        hkernel: 卷积核高度
        wkernel: 卷积核宽度
        hpad: 填充高度
        wpad: 填充宽度
        hstride: 步幅高度
        wstride: 步幅宽度
        groups: 组卷积个数。默认 1。暂未实现组卷积
        dilation: 空洞卷积参数。默认 (1, 1)。暂时不支持空洞卷积
    """
    batch: int
    height: int
    width: int
    in_channels: int
    out_channels: int
    hkernel: int
    wkernel: int
    hpad: int
    wpad: int
    hstride: int
    wstride: int
    groups: int = 1  # 暂未实现组卷积
    dilation: tuple[int, int] = (1, 1)  # 暂时不支持空洞卷积

    @property
    def fout_height(self):
        """输出高度"""
        return conv_out_size(self.height, self.hkernel, self.hpad, self.hstride)

    @property
    def fout_width(self):
        """输出宽度"""
        return conv_out_size(self.width, self.wkernel, self.wpad, self.wstride)

    @property
    def fout_shape(self):
        """输出形状"""
        return self.batch, self.out_channels, self.fout_height, self.fout_width

    def gen_data(self, data_dtype="float32",
                 kernel_dtype="float32"):
        """模拟卷积的参数"""
        data_np = np.random.normal(size=(self.batch, self.in_channels,
                                   self.height, self.width)).astype(data_dtype)
        kernel_np = np.random.normal(size=(self.out_channels, self.in_channels,
                                     self.hkernel, self.wkernel)).astype(kernel_dtype)
        return data_np, kernel_np

    def __call__(self, data_np, kernel_np, out_dtype="float32"):
        """计算卷积"""
        ref_np = np.empty(self.fout_shape, dtype=out_dtype)
        pad_data_np = pad2d(data_np, self.hpad, self.wpad)
        for b in range(self.batch):
            for c_o in range(self.out_channels):
                for i in range(self.fout_height):
                    for j in range(self.fout_width):
                        ref_np[b, c_o, i, j] = 0  # 初始化
                        for c_i in range(self.in_channels):
                            for h_k in range(self.hkernel):
                                for w_k in range(self.wkernel):
                                    _x = pad_data_np[b, c_i, i *
                                                     self.hstride + h_k, j*self.wstride+w_k]
                                    ref_np[b, c_o, i, j] += kernel_np[c_o,
                                                                      c_i, h_k, w_k] * _x
        return ref_np



import time
import mxnet as mx
from gluoncv.model_zoo import get_model


def params_count(net, input_size):
    _params_count = 0
    net.initialize()
    _ = net(mx.nd.zeros((1, 3, input_size, input_size)))
    for _, v in net.collect_params().items():
        _params_count += v.data().size
    return _params_count


def load_deploy_model(model_prefix: str):
    '''
    Args:
        model_prefix: load static model as hybridblock.
    '''
    net = mx.gluon.SymbolBlock.imports(f'{model_prefix}-symbol.json',
                                       ['data'], f'{model_prefix}-0000.params')
    net.hybridize(static_alloc=True, static_shape=True)
    return net


def load_model(model_name, ctx, classes,
               pretrained: bool = True,
               dtype: str = 'float32',
               quantized: bool = False,
               use_se: bool = False,
               **kwargs):
    '''

    Args:
        dtype: training data type
        params_file: local parameter file to load, instead of pre-trained weight.
        quantized: use int8 pretrained model
        use_se: use SE layers or not in resnext. default is false.
    '''
    kwargs = {'pretrained': pretrained,
              'ctx': ctx,
              'classes': classes}
    if model_name.startswith('resnext'):
        kwargs['use_se'] = use_se
    net = get_model(model_name, **kwargs)
    if not pretrained:
        net.initialize()
    net.cast(dtype)
    if quantized:
        net.hybridize(static_alloc=True, static_shape=True)
    else:
        net.hybridize()
    return net


def load_model2(model_name, ctx, classes,
                dtype: str = 'float32',
                quantized: bool = False,
                params_file='',
                use_se: bool = False,
                **kwargs):
    '''

    Args:
        dtype: training data type
        params_file: local parameter file to load, instead of pre-trained weight.
        quantized: use int8 pretrained model
        use_se: use SE layers or not in resnext. default is false.
    '''
    pretrained = True if not params_file else False
    kwargs = {'pretrained': pretrained,
              'ctx': ctx,
              'classes': classes}
    if model_name.startswith('resnext'):
        kwargs['use_se'] = use_se
    net = get_model(model_name, **kwargs)
    net.cast(dtype)
    if params_file:
        net.load_parameters(params_file, ctx=ctx)
    if quantized:
        net.hybridize(static_alloc=True, static_shape=True)
    else:
        net.hybridize()
    return net


def benchmark(network, ctx,
              batch_size=64,
              image_size=224,
              num_iter=100,
              dtype='float32'):
    '''
    Examples:
        print('-----benchmark mode for model %s-----' % model_name)
        time_cost = benchmark(network=net, ctx=ctx[0], image_size=input_size, batch_size=batch_size,
                            num_iter=num_batches, datatype='float32')
        fps = (batch_size*num_batches)/time_cost
        print('With batch size %s, %s batches, inference performance is %.2f img/sec' %
            (batch_size, num_batches, fps))
    '''
    input_shape = (batch_size, 3) + (image_size, image_size)
    data = mx.random.uniform(-1.0, 1.0, shape=input_shape,
                             ctx=ctx, dtype=dtype)
    dryrun = 5
    for i in range(num_iter+dryrun):
        if i == dryrun:
            tic = time.time()
        output = network(data)
        output.asnumpy()
    toc = time.time() - tic
    return toc


def get_fps(batch_size, num_batches, time_cost):
    return (batch_size*num_batches)/time_cost



from dataclasses import dataclass
import numpy as np


@dataclass(eq=True, order=True)
class TensorType:
    shape: tuple[int]
    dtype: str
    name: str = "data"

    @property
    def empty(self):
        '''空的输出张量'''
        return np.empty(shape=self.shape,
                        dtype=self.dtype)

    @property
    def nbytes(self):
        '''输出张量的存储字节数'''
        return self.empty.nbytes

    @property
    def nelement(self):
        '''输出张量的元素个数'''
        return self.empty.size
